import 'dart:convert';
import 'package:http/http.dart' as http;
import 'package:logger/logger.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:xml2json/xml2json.dart';
import 'package:html/parser.dart' as parser;

/// Authentication methods supported by the API
enum AuthMethod { none, cookie, basic }

/// Authentication systems for proxies
enum AuthSystem {
  none,
  authelia,
  cloudflareZeroTrust,
  swag,
  traefik,
  nginxProxy,
  custom,
}

/// Service class to handle API requests with various authentication methods
class ApiService {
  final Logger _logger = Logger();
  final http.Client _client = http.Client();
  String? _baseUrl;
  String? _cookie;
  String? _username;
  String? _password;
  String? _basePath;
  AuthSystem _authSystem = AuthSystem.none;

  static final ApiService _instance = ApiService._internal();
  factory ApiService() => _instance;

  ApiService._internal();

  /// Returns the base URL or an empty string
  String getBaseUrl() {
    return _baseUrl ?? '';
  }

  /// Returns the username or an empty string
  String getUsername() {
    return _username ?? '';
  }

  /// Returns the password or an empty string
  String getPassword() {
    return _password ?? '';
  }

  /// Initializes the API service with credentials from shared preferences
  Future<void> initialize() async {
    final prefs = await SharedPreferences.getInstance();
    _baseUrl = prefs.getString('base_url');
    _cookie = prefs.getString('calibre_web_session');
    _username = prefs.getString('username');
    _password = prefs.getString('password');
    _basePath = prefs.getString('base_path') ?? '';

    final authSystemString = prefs.getString('auth_system') ?? 'none';
    try {
      _authSystem = AuthSystem.values.firstWhere(
        (e) => e.toString().split('.').last == authSystemString,
        orElse: () => AuthSystem.none,
      );
    } catch (e) {
      _authSystem = AuthSystem.none;
    }

    _logger.i('Initialized API service with auth system: $_authSystem');
  }

  void dispose() {
    _client.close();
  }

  /// Makes an authenticated GET request
  /// Returns the parsed JSON response or throws an exception
  ///
  /// Parameters:
  ///
  /// - `endpoint`: The API endpoint to request
  /// - `authMethod`: The authentication method to use
  /// - `queryParams`: Optional query parameters
  Future<Map<String, dynamic>> getJson(
    String endpoint,
    AuthMethod authMethod, {
    Map<String, String>? queryParams,
  }) async {
    final response = await get(endpoint, authMethod, queryParams: queryParams);
    try {
      if (response.body.length > 50) {
        _logger.d('Response body: ${response.body.substring(0, 50)}...');
      } else {
        _logger.d('Response body: ${response.body}');
      }
      return json.decode(response.body) as Map<String, dynamic>;
    } catch (e) {
      _logger.e('Failed to parse JSON response: $e');

      _logger.d('Response body: ${response.body}...');
      throw FormatException('Invalid JSON response: $e');
    }
  }

  /// Makes an authenticated GET request
  /// Returns the raw response object
  ///
  /// Parameters:
  ///
  /// - `endpoint`: The API endpoint to request
  /// - `authMethod`: The authentication method to use
  /// - `queryParams`: Optional query parameters
  Future<http.Response> get(
    String endpoint,
    AuthMethod authMethod, {
    Map<String, String>? queryParams,
  }) async {
    await _ensureInitialized();
    final uri = _buildUri(endpoint, queryParams);
    final headers = _getAuthHeaders(authMethod);

    // Add processed custom headers for auth system
    final customHeaders = await _processCustomHeaders();
    headers.addAll(customHeaders);

    _logger.d('GET request to: $uri');
    _logger.d(
      'Using ${authMethod.name} authentication with ${_authSystem.name} proxy system',
    );
    _logger.d('Headers: $headers');

    try {
      final response = await _client.get(uri, headers: headers);
      _logger.i('Response status: ${response.statusCode}');
      _checkResponseStatus(response.statusCode);
      return response;
    } catch (e) {
      _logger.e('Request failed: $e');
      rethrow;
    }
  }

  /// Makes an authenticated POST request with optional CSRF token
  ///
  /// Parameters:
  ///
  /// - `endpoint`: The API endpoint to request
  /// - `queryParams`: Optional query parameters
  /// - `body`: The request body
  /// - `authMethod`: The authentication method to use
  /// - `contentType`: The content type of the request
  /// - `useCsrf`: Whether to fetch and include CSRF token
  /// - `csrfSelector`: CSS selector for the CSRF token input field
  /// Makes an authenticated POST request with optional CSRF token
  ///
  /// Parameters:
  ///
  /// - `endpoint`: The API endpoint to request
  /// - `queryParams`: Optional query parameters
  /// - `body`: The request body
  /// - `authMethod`: The authentication method to use
  /// - `contentType`: The content type of the request
  /// - `useCsrf`: Whether to fetch and include CSRF token
  /// - `csrfSelector`: CSS selector for the CSRF token input field
  Future<http.Response> post(
    String endpoint,
    Map<String, String>? queryParams,
    dynamic body,
    AuthMethod authMethod, {
    String contentType = 'application/json',
    bool useCsrf = false,
    String csrfSelector = 'input[name="csrf_token"]',
  }) async {
    await _ensureInitialized();
    final uri = _buildUri(endpoint, queryParams);

    // Add processed custom headers for auth system
    final customHeaders = await _processCustomHeaders();

    // If we need to handle CSRF protection, use a two-step process
    if (useCsrf) {
      _logger.i('Making CSRF-protected POST request to: $uri');

      // STEP 1: Make initial GET request to fetch CSRF token
      final getHeaders = _getAuthHeaders(authMethod);
      getHeaders.addAll(customHeaders);
      getHeaders['Accept'] = 'text/html,application/xhtml+xml,application/xml';

      final getResponse = await _client.get(uri, headers: getHeaders);
      _logger.d(
        'GET response status for CSRF fetch: ${getResponse.statusCode}',
      );

      if (getResponse.statusCode != 200) {
        _logger.e(
          'Initial GET request for CSRF token failed: ${getResponse.statusCode}',
        );
        throw Exception(
          'Failed to fetch CSRF token: ${getResponse.statusCode}',
        );
      }

      // Extract CSRF token from HTML
      final document = parser.parse(getResponse.body);
      final csrfElement = document.querySelector(csrfSelector);
      final csrfToken = csrfElement?.attributes['value'];

      if (csrfToken == null) {
        _logger.e('Could not find CSRF token using selector: $csrfSelector');
        throw Exception('CSRF token not found');
      }

      // Extract new session cookie if available
      String sessionCookie = _cookie ?? '';
      if (getResponse.headers.containsKey('set-cookie')) {
        final setCookieHeader = getResponse.headers['set-cookie']!;
        final sessionMatch = RegExp(
          r'session=([^;]+)',
        ).firstMatch(setCookieHeader);
        if (sessionMatch != null && sessionMatch.groupCount >= 1) {
          sessionCookie = 'session=${sessionMatch.group(1)}';
        }
      }

      // STEP 2: Make POST request with extracted CSRF token
      final postHeaders = {
        'Content-Type': contentType,
        'Cookie': sessionCookie,
        'X-CSRFToken': csrfToken,
        'X-Requested-With': 'XMLHttpRequest',
        'Referer': uri.toString(),
        'Origin':
            '${uri.scheme}://${uri.host}${uri.port != 80 && uri.port != 443 ? ":${uri.port}" : ""}',
      };
      postHeaders.addAll(customHeaders);

      // Add CSRF token to body if it's a map
      Map<String, dynamic> finalBody;
      if (body is Map) {
        if (body is Map<String, dynamic>) {
          finalBody = Map<String, dynamic>.from(body);
        } else {
          finalBody = Map<String, dynamic>.from(
            body.map((key, value) => MapEntry(key.toString(), value)),
          );
        }
        finalBody['csrf_token'] = csrfToken;
      } else {
        finalBody = {'csrf_token': csrfToken};
      }

      final encodedBody = _encodeBody(finalBody, contentType);

      _logger.d('CSRF-protected POST headers: $postHeaders');
      _logger.d('CSRF-protected POST body: $encodedBody');

      try {
        final response = await _client.post(
          uri,
          headers: postHeaders,
          body: encodedBody,
        );
        _logger.i(
          'CSRF-protected POST response status: ${response.statusCode}',
        );
        _checkResponseStatus(response.statusCode);
        return response;
      } catch (e) {
        _logger.e('CSRF-protected POST request failed: $e');
        rethrow;
      }
    } else {
      // Standard POST request without CSRF protection
      final headers = _getAuthHeaders(authMethod);
      headers['Content-Type'] = contentType;
      headers.addAll(customHeaders);

      _logger.d('POST request to: $uri');
      _logger.d(
        'Using ${authMethod.name} authentication with ${_authSystem.name} proxy system',
      );
      _logger.d('Headers: $headers');

      final encodedBody = _encodeBody(body, contentType);

      try {
        final response = await _client.post(
          uri,
          headers: headers,
          body: encodedBody ?? "",
        );
        _logger.i('POST response status: ${response.statusCode}');
        _checkResponseStatus(response.statusCode);
        return response;
      } catch (e) {
        _logger.e('POST request failed: $e');
        rethrow;
      }
    }
  }

  /// Makes an authenticated GET request and returns a StreamedResponse
  /// This is useful for downloading files or streaming large responses
  ///
  /// Parameters:
  ///
  /// - `endpoint`: The API endpoint to request
  /// - `authMethod`: The authentication method to use
  Future<http.StreamedResponse> getStream(
    String endpoint,
    AuthMethod authMethod,
  ) async {
    await _ensureInitialized();

    // Ensure URL is complete with base path
    String fullPath = endpoint;
    if (_basePath != null &&
        _basePath!.isNotEmpty &&
        !endpoint.startsWith('http')) {
      if (endpoint.startsWith('/')) {
        fullPath = '/$_basePath$endpoint';
      } else {
        fullPath = '/$_basePath/$endpoint';
      }
    }
    final fullUrl =
        endpoint.startsWith('http') ? endpoint : '$_baseUrl$fullPath';

    final headers = _getAuthHeaders(authMethod);

    // Add processed custom headers for auth system
    final customHeaders = await _processCustomHeaders();
    headers.addAll(customHeaders);

    _logger.d('GET stream request to: $fullUrl');
    _logger.d(
      'Using ${authMethod.name} authentication with ${_authSystem.name} proxy system',
    );
    _logger.d('Headers: $headers');

    final request = http.Request('GET', Uri.parse(fullUrl));
    request.headers.addAll(headers);

    try {
      final response = await _client.send(request);
      _logger.i('Stream response status: ${response.statusCode}');
      _checkResponseStatus(response.statusCode);
      return response;
    } catch (e) {
      _logger.e('Stream request failed: $e');
      rethrow;
    }
  }

  /// Fetches a CSRF token from the specified endpoint
  ///
  /// Parameters:
  ///
  /// - `endpoint`: The endpoint to fetch the token from
  /// - `authMethod`: The authentication method to use
  /// - `selector`: CSS selector for the CSRF token input
  Future<Map<String, String?>> fetchCsrfToken(
    String endpoint,
    AuthMethod authMethod,
    String selector,
  ) async {
    _logger.d('Fetching CSRF token from: $endpoint');
    final response = await get(endpoint, authMethod);

    final document = parser.parse(response.body);
    final csrfElement = document.querySelector(selector);
    final csrfToken = csrfElement?.attributes['value'];

    if (csrfToken == null) {
      _logger.w(
        'CSRF token not found in the response using selector: $selector',
      );
    } else {
      _logger.d('CSRF token found: $csrfToken');
    }

    return {'token': csrfToken, 'cookies': response.headers['set-cookie']};
  }

  /// Ensures credentials are loaded before making requests
  Future<void> _ensureInitialized() async {
    if (_baseUrl == null) {
      await initialize();

      if (_baseUrl == null) {
        throw Exception(
          'Server URL is missing. Please configure the app settings.',
        );
      }
    }
  }

  /// Builds a URI for API requests
  ///
  /// Parameters:
  ///
  /// - `endpoint`: The API endpoint to request
  Uri _buildUri(String endpoint, Map<String, String>? queryParams) {
    // Handle base path if set
    String fullPath = endpoint;
    if (_basePath != null && _basePath!.isNotEmpty) {
      // Make sure we don't duplicate slashes
      if (endpoint.startsWith('/')) {
        fullPath = '/$_basePath$endpoint';
      } else {
        fullPath = '/$_basePath/$endpoint';
      }
    }

    return Uri.parse(
      '$_baseUrl$fullPath',
    ).replace(queryParameters: queryParams);
  }

  /// Process custom headers, replacing placeholders with actual values
  Future<Map<String, String>> _processCustomHeaders() async {
    final prefs = await SharedPreferences.getInstance();
    final headersJson = prefs.getString('custom_login_headers') ?? '[]';

    final List<dynamic> decodedList = jsonDecode(headersJson);
    final List<Map<String, String>> customHeaders =
        decodedList
            .map((item) => Map<String, String>.from(item as Map))
            .toList();

    Map<String, String> processedHeaders = {};

    for (var header in customHeaders) {
      String key = header.keys.first;
      String value = header.values.first;

      // Replace username placeholder if available
      if (value.contains('\${USERNAME}') && _username != null) {
        value = value.replaceAll('\${USERNAME}', _username!);
      }

      processedHeaders[key] = value;
    }

    return processedHeaders;
  }

  /// Gets authentication headers based on the auth method
  ///
  /// Parameters:
  ///
  /// - `authMethod`: The authentication method to use
  Map<String, String> _getAuthHeaders(AuthMethod authMethod) {
    Map<String, String> headers = {};

    if (authMethod == AuthMethod.cookie && _cookie != null) {
      headers['Cookie'] = _cookie!;
    } else if (_username != null && _password != null) {
      headers['Authorization'] =
          'Basic ${base64.encode(utf8.encode('$_username:$_password'))}';
    }

    return headers;
  }

  /// Encodes request body based on content type
  ///
  /// Parameters:
  ///
  /// - `body`: The request body to encode
  /// - `contentType`: The content type of the request
  dynamic _encodeBody(dynamic body, String contentType) {
    if (body is Map) {
      if (contentType == 'application/json') {
        return json.encode(body);
      } else if (contentType == 'application/x-www-form-urlencoded') {
        // Convert map to URL encoded string format key1=value1&key2=value2
        return body.entries
            .map(
              (e) =>
                  '${Uri.encodeComponent(e.key.toString())}=${Uri.encodeComponent(e.value.toString())}',
            )
            .join('&');
      }
    }
    return body;
  }

  /// Checks response status code and throws appropriate exceptions
  ///
  /// Parameters:
  ///
  /// - `statusCode`: The status code to check
  void _checkResponseStatus(int statusCode) {
    if (statusCode == 401) {
      throw Exception('Authentication failed. Please log in again.');
    } else if (statusCode >= 500) {
      throw Exception('Server error: $statusCode');
    } else if (statusCode >= 400) {
      throw Exception('Request failed with status $statusCode');
    }
  }

  /// Makes an authenticated GET request and converts XML response to JSON using Parker format
  /// Returns the parsed JSON response or throws an exception
  ///
  /// Parameters:
  ///
  /// - `endpoint`: The API endpoint to request
  /// - `authMethod`: The authentication method to use
  /// - `queryParams`: Optional query parameters
  Future<Map<String, dynamic>> getXmlAsJson(
    String endpoint,
    AuthMethod authMethod, {
    Map<String, String>? queryParams,
  }) async {
    final response = await get(endpoint, authMethod, queryParams: queryParams);
    try {
      // XML to JSON conversion
      final transformer = Xml2Json();
      transformer.parse(response.body);

      String jsonString = transformer.toParkerWithAttrs();

      return json.decode(jsonString) as Map<String, dynamic>;
    } catch (e) {
      _logger.e('Failed to parse XML response to JSON: $e');
      final previewLength =
          response.body.length > 100 ? 100 : response.body.length;
      _logger.d(
        'Response body: ${response.body.substring(0, previewLength)}...',
      );
      throw FormatException('Invalid XML response: $e');
    }
  }
}
